#!/usr/bin/env python

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import re

from PyQt6 import QtWidgets, QtGui
from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

from .EdView import *

# inherit from EditView, add editing support.
class EditorViewRW(EditorView):
	sig_file_modified = pyqtSignal(bool)
	def __init__(self, parent=None):
		EditorView.__init__(self, parent)
		# use default settings.
		EditorView.ed_settings_1(self)

	def open_file(self, filename):
		self.open_file_begin(filename)

		## Show this file in the editor
		data = open(filename).read()
		try:
			data = data.decode("UTF-8")
		except:
			pass
		self.setText(data)

		## process for modifiled.
		self.setModified(False)
		self.modificationChanged.connect(self.modifiedChanged)
		
		## support edit 
		self.setReadOnly(False)

		self.open_file_end()

	def refresh_file(self, filename):
		if self.isModified():
			msg = 'Save changes before refresh ?'
			if DialogManager.show_yes_no(msg):
				self.save_file(self.filename)
		EditorView.refresh_file(self, filename)

	def modifiedChanged(self):
		self.sig_file_modified.emit(self.isModified())
	
	# FIXME: too simple for big files or remote files.	
	def save_file(self, filename):
		if(self.isModified()):
			fobj = open(filename, 'w')
			if (not fobj.closed):
				fobj.write(self.text())
				fobj.flush()
				fobj.close()
				self.setModified(False)

class EditorPageRW(EditorPage):
	def new_editor_view(self):
		return EditorViewRW(self)


class EditorBookRW(EditorBook):

	def new_editor_page(self):
		return EditorPageRW()

	def save_file_at_inx(self, inx):
		if inx < 0:
			return
		page = self.widget(inx)
		filename = page.get_filename()
		if filename:
			page.ev.save_file(filename)
			page.fcv.rerun(filename)

	def save_current_page(self):
		inx = self.currentIndex()
		self.save_file_at_inx(inx)

	def save_tab_list(self, inx_list):
		for inx in inx_list:
			self.save_file_at_inx(inx)

	def save_all_file(self):
		inx_list = list(range(self.count()))
		self.save_tab_list(inx_list)

	def page_modified_cb(self, isModifiled):
		inx = self.currentIndex()
		filename = self.tabText(inx)
		# Sign modified.
		if isModifiled:
			self.tabBar().setTabTextColor(inx, Qt.GlobalColor.red)
		else:
			self.tabBar().setTabTextColor(inx, Qt.GlobalColor.black)
	
	def close_cb(self, inx):
		page = self.widget(self.currentIndex())
		if page.ev.isModified():
			if DialogManager.show_yes_no("Do you want to save file ?"):
				self.save_current_page()
		self.removeTab(inx)

	def has_modified_file(self, inx_list):
		for i in inx_list:
			page = self.widget(i)
			if page.ev.isModified():
				return True
		return False

	def close_list_common(self, type):
		inx_list = self.tab_list(self.currentIndex(), type)
		if len(inx_list) == 0:
			return
		if self.has_modified_file(inx_list):
			msg = 'Closing all %s.\nSave changes ?' % type
			if DialogManager.show_yes_no(msg):
				self.save_tab_list()
			self.remove_tab_list(inx_list)
		else:
			msg = 'Close all %s ?' % type
			if not DialogManager.show_yes_no(msg):
				return
			self.remove_tab_list(inx_list)

	def show_file_line(self, filename, line, hist=True):
		EditorBook.show_file_line(self, filename, line, hist=hist)
		page = self.currentWidget()
		# modified signal callback
		page.ev.sig_file_modified.connect(self.page_modified_cb)
		
	# redo editing callback
	def redo_edit_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.redo()

	# undo editing callback
	def undo_edit_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.undo()

	# edting callbacks
	def cut_edit_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.cut()

	def paste_edit_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.paste()

	



#!/usr/bin/env python

# Copyright (c) 2014 Anil Kumar
# All rights reserved.
#
# License: BSD

import sys
sys.dont_write_bytecode = True

import os
import socket

import BackendTool
import SeascopeServer

def start(plist, host, port):
	if not BackendTool.load_projects(plist):
		sys.exit(-1)

	SeascopeServer.start_server(host, port)

if __name__ == '__main__':
	import optparse
	usage = "usage: %prog [options] [project_path ...]"
	op = optparse.OptionParser(usage=usage)
	op.add_option("-H", "--host", dest="host",             default="127.0.0.1", help="Host")
	op.add_option("-P", "--port", dest="port", type='int', default=8650,        help="Port")
	(options, args) = op.parse_args()

	if options.port > 65535:
		print('options.port > 65535')
		sys.exit(-1)

	plist = []
	for p in args:
		if not os.path.isdir(p):
			print('%p is not a directory')
			sys.exit(-2)
		plist.append(os.path.abspath(p))

	#options.host = socket.gethostname()
	start(plist, options.host, options.port)



#!/usr/bin/env python

# Copyright (c) 2014 Anil Kumar
# All rights reserved.
#
# License: BSD

import sys
import os
import subprocess
import threading
import time

import json
import socket
import signal

import BackendTool

from socketserver import ThreadingMixIn
from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler

server_conf_js = None

def setup_srvr_conf_js(host, port):
	global server_conf_js
	server_conf_js = '''var sinfo = {
				host: "%s",
				port: %s,
			}''' % (host, port)

class ThreadingServer(ThreadingMixIn, HTTPServer):
	allow_reuse_address = True
	daemon_threads = True

is_debug = os.getenv('SEASCOPE_WEB_DEBUG')

class RequestHandler(SimpleHTTPRequestHandler):
	def parse_path(self):
		import urllib.parse
		p = urllib.parse.urlparse(self.path)
		if is_debug:
			print('path', p.path)
		if p.path.startswith('/'):
			path = p.path
			if path == '/':
				path = '/h.html'
			if path == '/js/server_conf.js':
				self.reply_to_GET(server_conf_js, 'text/javascript');
				return None
			wp = 'www' + path
			if os.path.exists(wp):
				data = open(wp).read()
				cont_type = 'text/html'
				if p.path.endswith('.js'):
					cont_type = 'text/javascript'
				self.reply_to_GET(data, cont_type);
				return None
		if p.path != '/q':
			if is_debug:
				print('not /q instead', p.path)
			return
		if is_debug:
			print('p.query', p.query)
		qd = urllib.parse.parse_qs(p.query)
		for k in list(qd.keys()):
			v = qd[k]
			if len(v) == 0:
				print('len(v) == 0', k, v)
				continue
			qd[k] = v[-1]
		if is_debug:
			print('qd', qd)
		return qd

	def reply_to_GET(self, qres, cont_type):
		rh = self
		rh.send_response(200)
		qres = qres.encode()
		#rh.send_header('Content-type', 'text/plain')
		rh.send_header('Content-type', cont_type)
		rh.send_header('Content-length', len(qres))
		rh.end_headers()
		rh.wfile.write(qres)

	def do_GET(self):
		qd = self.parse_path()
		if not qd:
			return
		outd = BackendTool.run_op(qd)
		try:
			data = json.dumps(outd, ensure_ascii=False)
			if 'callback' in qd:
				data = '%s(%s);' % (qd['callback'], data)
			self.reply_to_GET(data, 'application/json')
		except Exception as e:
			print('qd', qd)
			print('outd', outd)
			print(e)
			data = e
			return

def start_server(host, port):
	setup_srvr_conf_js(host, port)

	print('starting server at http://%s:%s' % (host, port))
	try:
		ts = ThreadingServer((host, port), RequestHandler)

		def signal_handler(signal, frame):
			ts.socket.close()
			print('got signal %s, exiting' % signal)
			os._exit(0)
		signal.signal(signal.SIGINT, signal_handler)
		signal.signal(signal.SIGTERM, signal_handler)

		ts.serve_forever()
	except Exception as e:
		print(e)
		sys.exit(-1)

	print('exiting')



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import re, os

from PyQt6 import QtWidgets, QtGui, QtCore, uic

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

fc_plugins = []
fc_dict = {}

def _load_plugins(module, directory):
	pluginImports = __import__(module, globals(), locals())
	print('Scanning for view plugins...')
	plist = []
	pdict = {}
	for i in sorted(os.listdir(directory)):
		path = os.path.join( directory, i, '__init__.py' )
		if os.path.isfile( path ):
			p = __import__( '%s.%s' % (module, i), globals(), locals(), ['*'] )
			plist.append(p)
			pdict[p.name()] = p
			if not hasattr(p, 'priority'):
				p.priority = 0 
	plist = sorted(plist, key=lambda p: p.priority, reverse=True)
	for p in plist:
		print('\t', p.name())
	return (plist, pdict)

def load_plugins():
	global fc_plugins, fc_dict
	(fc_plugins, fc_dict) = _load_plugins('view.filecontext.plugins', 'view/filecontext/plugins')

def run_plugins(filename, parent):
	for p in fc_plugins:
		if not hasattr(p, 'cmd_name'):
			p.run_plugin(filename, parent)




#!/usr/bin/env python3

if __name__ == '__main__':
	import sys, os, optparse
	
	usage = "usage: %prog filename"
	op = optparse.OptionParser(usage=usage)
	(options, args) = op.parse_args()

	if len(args) != 1:
		print('Please specify a filename', file=sys.stderr)
		sys.exit(-1)
	filename = args[0]
	filename = os.path.abspath(filename)
	if not os.access(filename, os.R_OK):
		print('Failed to read the file:', filename)
		sys.exit(-2)


	app_dir = os.path.dirname(os.path.realpath(__file__))
	os.chdir(app_dir)

	from PyQt6 import QtWidgets, QtGui, QtCore
	from PyQt6.QtWidgets import *
	from PyQt6.QtGui import *
	from PyQt6.QtCore import *
	from view import EdView, EdViewRW
	import view
	
	view.load_plugins()

	app = QApplication(sys.argv)

	w = QMainWindow()
	if os.getenv("SEASCOPE_EDIT", 0):
		book = EdViewRW.EditorBookRW()
	else:
		book = EdView.EditorBook()

	actDescr = [
		[ book.find_cb,      'Ctrl+F' ],
                [ book.find_next_cb, 'F3' ],
                [ book.find_prev_cb, 'Shift+F3' ],
        ]
	for ad in actDescr:
		act = QAction(w)
		act.setShortcut(ad[1])
		act.triggered.connect(ad[0])
		w.addAction(act)

	w.setCentralWidget(book)
	w.resize(900, 600)
	w.show()

	book.show_file(filename)
	w.setWindowTitle(filename)

	sys.exit(app.exec())



#!/usr/bin/env python3

import os, sys, re
import subprocess
import shutil
import tempfile

GRAPH_ON_TYPES = ['function', 'macro']

def _eintr_retry_call(func, *args):
	while True:
		try:
			return func(*args)
		except OSError as e:
			if e.errno == errno.EINTR:
				continue
			raise

def ct_override(filename):
	pattern = os.getenv('SEASCOPE_CTAGS_OVERRIDE_PATTERN')
	delim1  = os.getenv('SEASCOPE_CTAGS_OVERRIDE_DELIM_START')
	delim2  = os.getenv('SEASCOPE_CTAGS_OVERRIDE_DELIM_END')
	if not (pattern and delim1 and delim2):
		return

	args = ['grep', '-En', "%s" % pattern, "%s" % filename]
	try:
		proc = subprocess.Popen(args, stdout=subprocess.PIPE)
		(out_data, err_data) = _eintr_retry_call(proc.communicate)
		out_data = out_data.decode()
		out_data = out_data.split('\n')
	except Exception as e:
		out_data =  [
				'Failed to run ct override cmd\tignore\t0;\t ',
				'cmd: %s\tignore\t0;\t ' % ' '.join(args),
				'error: %s\tignore\t0;\t ' % str(e),
				'ctags not installed ?\tignore\t0;\t ',
			]
	res = {}
	for line in out_data:
		if (line == ''):
			break
		line = line.split(':', 1)
		num = line[0]
		inx1 = line[1].find(delim1) + 1
		inx2 = line[1].find(delim2)
		sym = line[1][inx1:inx2]
		res[num] = [sym, num, 'function']
	return res


def ct_cmdForFile(f):
	#ct_args = 'ctags -n -u --fields=+K -f - --extra=+q'
	#ct_args = 'ctags -n -u --fields=+Ki -f -'
	ct_args = 'ctags -n -u --fields=+K -f -'
	opt_I_file = os.getenv('SEASCOPE_CTAGS_OPT_I_FILE')
	if opt_I_file and os.path.isfile(opt_I_file):
		ct_args += ' -I ' + opt_I_file
	if os.path.isdir(f):
		cmd = ct_args + ' -R'
		return cmd
	suffix_cmd_map = []
	custom_map = os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP')
	if custom_map:
		custom_map = eval(custom_map)
		suffix_cmd_map += custom_map
	suffix_cmd_map.append( ['', ct_args] )
	for (suffix, cmd) in suffix_cmd_map:
		if f.endswith(suffix):
			return cmd
	return None

def ct_query(filename):
	override_res = ct_override(filename)

	args = ct_cmdForFile(filename)
	args = args.split()
	args.append(filename)
	try:
		proc = subprocess.Popen(args, stdout=subprocess.PIPE)
		(out_data, err_data) = _eintr_retry_call(proc.communicate)
		out_data = out_data.decode()
		out_data = out_data.split('\n')
	except Exception as e:
		out_data =  [
				'Failed to run ctags cmd\tignore\t0;\t ',
				'cmd: %s\tignore\t0;\t ' % ' '.join(args),
				'error: %s\tignore\t0;\t ' % str(e),
				'ctags not installed ?\tignore\t0;\t ',
			]
	res = []
	for line in out_data:
		if (line == ''):
			break
		line = line.split('\t')
		num = line[2].split(';', 1)[0]
		if override_res:
			override_ent = override_res.pop(num, None)
			if override_ent:
				line[0] = override_ent[0]
		line = [line[0], num, line[3]]
		res.append(line)
	if override_res:
		res += override_res.values()
		res = sorted(res, key=lambda e: int(e[1]))
	return res

class FFgraph:
	def __init__(self):
		self.cs_cmd_base = None

	def parse_cs_result(self, text):
		text = re.split(r'\r?\n', text)
		res = []
		for line in text:
			if line == '':
				continue
			line = line.split(' ', 3)
			line = [line[1], line[0], line[2], line[3]]
			res.append(line)
		return res

	def findCalleeList(self, caller, srcFile):
		try:
			cmd = self.cs_cmd_base + [ '-2', caller ]
			if os.path.isdir(srcFile):
				cmd = cmd + [ '-s' ]
			cmd = cmd + [ srcFile ] 
			# In python >= 2.7 can use subprocess.check_output
			# output = subprocess.check_output(cmd)
			proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
			(output, err_data) = proc.communicate()
			output = output.decode()
			#print 'cmd =', cmd
			#print 'output =', output
			res = self.parse_cs_result(output)
			res = set([ e[0] for e in res ])
		except Exception as e:
			print('failed cmd:', cmd, ':', e, file=sys.stderr)
			res = []
		return res

	def generateDotHeader(self, name):
		out = 'digraph "%s" {\n' % name
		out += 'rankdir="LR"\n'
		#out += '\tconcentrate=true;\n'
		return out

	def generateDotFooter(self):
		out = '}\n\n'
		return out
		
	def generateSubgraphHeader(self, name):
		out = '\tsubgraph "%s" {\n' % name
		return out

	def generateSubgraphFooter(self):
		out = '\t}\n\n'
		return out

	def generateClusterInfoLocal(self, localSym):
		out = ''
		# local
		out += self.generateSubgraphHeader('cluster_local_sym')
		for sym in localSym:
			out += '\t\t%s;\n' % sym
		out += self.generateSubgraphFooter()
		return out

	def generateClusterInfoExternal(self, externalSym):
		out = ''
		# extern
		out += self.generateSubgraphHeader('cluster_external_sym')
		for sym in externalSym:
			out += '\t\t%s;\n' % sym
		out += self.generateSubgraphFooter()
		return out

	def generateClusterInfo(self, localSym, externalSym):
		out = ''
		#out += self.generateClusterInfoLocal(localSym)
		out += self.generateClusterInfoExternal(externalSym)
		return out

	def generateDotInput(self, f, callerCalleInfo, is_extern):
		dotInput = ''
		dotInput += self.generateDotHeader(f)
		(localSym, externalSym, callerCalleInfo) = callerCalleInfo
		if is_extern:
			dotInput += self.generateClusterInfo(localSym, externalSym)
		# nodes
		for (caller, localCalleeList, externCalleeList) in callerCalleInfo:
			for callee in localCalleeList:
				dotInput += '\t"%s" -> "%s";\n' % (caller, callee)
			if not is_extern:
				continue
			for callee in externCalleeList:
				style = '[style=dotted]'
				dotInput += '\t"%s" -> "%s" %s;\n' % (caller, callee, style)
		dotInput += self.generateDotFooter()
		return dotInput

	def _generateDotGraph(self, f, is_extern, gOut):
		try:
			res = self.getCallerCalleeInfo(f, is_extern)
			dotInput = self.generateDotInput(f, res, is_extern)
		except Exception as e:
			print(e, file=sys.stderr)
		
		if not gOut:
			print(dotInput)
			return None

		args = ['dot', '-Tsvg']
		try:
			p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			(svg_data, err_data) = p.communicate(dotInput.encode())
			svg_data = svg_data.decode()
			if err_data and err_data.decode() != '':
				print(err_data, '\n', file=sys.stderr)
			#self.saveSvgFile(sym, svg_data)
			return svg_data
		except Exception as e:
			print('Failed to run:', ' '.join(args), '\n', file=sys.stderr)
			print('dotInput:', dotInput, '\n', file=sys.stderr)
			print(e, '\n', file=sys.stderr)
		

	def generateDotGraph(self, f, is_extern, gOut):
		tmp_cs_dir = tempfile.mkdtemp(prefix='ffgraph_')
		if not tmp_cs_dir:
			print('generateDotGraph: failed to create tmp dir')
			return
		self.cs_cmd_base = [ 'cscope', '-L', '-f', os.path.join(tmp_cs_dir, 'cscope.out') ]

		svg_data = None

		try:
			svg_data = self._generateDotGraph(f, is_extern, gOut)
		except:
			print('generateDotGraph: failed')

		shutil.rmtree(tmp_cs_dir)
		
		return svg_data

	def getCallerCalleeInfo(self, f, is_extern):
		ct_out = ct_query(f)
		localSym = dict([ (e[0], e[2]) for e in ct_out ])
		lSym = []
		eSym = []
		res = []
		visitedCaller = dict()
		for e in ct_out:
			if e[2] in GRAPH_ON_TYPES:
				caller = e[0]
				if caller in visitedCaller:
					continue
				visitedCaller[caller] = True
				calleeList = self.findCalleeList(caller, f)
				localCalleeList = []
				externCalleeList = []
				for c in calleeList:
					if c in localSym and localSym[c] in GRAPH_ON_TYPES:
						localCalleeList.append(c)
						continue
					if is_extern:
						externCalleeList.append(c)
				res.append([caller, localCalleeList, externCalleeList])
				if is_extern:
					if len(localCalleeList) or len(externCalleeList):
						lSym += [caller] + localCalleeList
						eSym += externCalleeList
				else:
					if len(localCalleeList):
						lSym += [caller] + localCalleeList
		return [lSym, eSym, res]

def ff_graph(f, is_extern, gOut=True):
	#if not os.path.isfile(f):
		#return None
	ffg = FFgraph()
	#gOut = False
	svg_data = ffg.generateDotGraph(f, is_extern, gOut)
	return svg_data

if __name__ == '__main__':
	import optparse
	usage = "usage: %prog [options] (-d <code_dir/file> | -p <idutils_proj>) (-e) [symbol]"
	op = optparse.OptionParser(usage=usage)
	op.add_option("-e", action="store_true", dest="is_extern", help="include external sym")
	op.add_option("-d", "--codedir", dest="code_dir", help="Code dir", metavar="CODE_DIR")
	op.add_option("-p", "--project", dest="id_path", help="Idutils project dir", metavar="PROJECT")
	(options, args) = op.parse_args()

	if (not any([options.code_dir, options.id_path]) or
		all([options.code_dir, options.id_path])):
		print('Specify one among -d or -p', file=sys.stderr)
		sys.exit(-1)

	sym = None
	dname = None
	is_extern = False
	if options.is_extern:
		is_extern = True
	if len(args):
		if len(args) != 1:
			print('Please specify only one symbol', file=sys.stderr)
			sys.exit(-2)
		sym = args[0]

	svg_data = None
	if options.code_dir:
		dname = options.code_dir
		if not os.path.exists(dname):
			print('"%s": does not exist' %  dname, file=sys.stderr)
			sys.exit(-3)
		svg_data = ff_graph(dname, is_extern=is_extern)
	print(svg_data)




#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import re

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

import DialogManager

class ResultPageItem(QTreeWidgetItem):
	def __init__(self, li):
		QTreeWidgetItem.__init__(self, li)
	def column_val(self, col):
		try:
			val = str(self.data(col, Qt.ItemDataRole.DisplayRole))
		except:
			return None
		return val

class ResultPage(QTreeWidget):
	sig_show_file_line = pyqtSignal(str, int)
	is_history_call = False

	def __init__(self, parent):
		QTreeWidget.__init__(self)

		self.is_history = False
		self.pbar = None
		self.parent = parent

		self.setColumnCount(4)
		self.cmd_items = ['Tag', 'File', 'Line', 'Text'] 
		self.setHeaderLabels(self.cmd_items)
		self.setColumnWidth(0, 200)
		self.setColumnWidth(1, 300)
		self.setColumnWidth(2, 40)

		# setup popup menu
		self.pmenu = QMenu()
		self.pmenu.addAction("&Filter", self.filter_cb)
		self.pmenu.addAction("&Show All", self.show_all_cb)
		self.pmenu.addSeparator()
		self.pmenu.addAction("&Remove Item", self.remove_item_cb)
		self.pmenu.addSeparator()
		self.pmenu.addAction("&Copy to Clipboard", self.copy_clipboard_cb)

		#self.setMinimumHeight(200)
		#self.setMinimumWidth(600)


		## set column width to fit contents
		##self.resizeColumnsToContents()
		##self.resizeRowsToContents()

		## set row height
		#nrows = len(result)
		#for row in xrange(nrows):
			#self.setRowHeight(row, 14)

		#self.setTextElideMode(Qt.TextElideMode.ElideLeft)
		self.setIndentation(-2)
		self.setAllColumnsShowFocus(True)
		
		self.activated.connect(self.activated_cb)

	def activated_cb(self, minx):
		(filename, line) = self.get_file_line(minx)
		if (not filename):
			return
		if (self.is_history):
			ResultPage.is_history_call = True
		self.sig_show_file_line.emit(filename, line)
		ResultPage.is_history_call = False

	def filter_cb(self):
		filtered = False
		cmd_sel = self.cmd_items[self.last_minx.column()]
		res = DialogManager.show_filter_dialog(self.cmd_items,cmd_sel)
		if res == None:
			return
		(filter_text, is_regex, is_negate, is_ignorecase, cmd_sel) = res
		cmd_idx = self.cmd_items.index(cmd_sel)
		if cmd_idx == -1:
			return
		for inx in range(self.topLevelItemCount()):
			item = self.topLevelItem(inx)
			text = item.column_val(cmd_idx)
			if (text == None):
				continue
			matched = False
			if (is_regex):
				if (is_ignorecase):
					if (re.search(filter_text, text, re.I) != None):
						matched = True
				else:
					if (re.search(filter_text, text) != None):
						matched = True
			else:
				if (is_ignorecase):
					if (text.upper().find(filter_text.upper()) != -1):
						matched = True
				else:
					if (text.find(filter_text) != -1):
						matched = True
			if (is_negate):
				matched = not matched
			if not matched:
				item.setHidden(True)
				filtered = True
		if filtered:
			self.mark_tab_filtered(True)

	def mark_tab_filtered(self, mark):
		tabIndex = self.parent.indexOf(self)
		label = self.parent.tabText(tabIndex)
		if mark and label[0] != "*":
			self.parent.setTabText(tabIndex, "*" + label)
		elif not mark and label[0] == "*":
			self.parent.setTabText(tabIndex, label[1:])

	def show_all_cb(self):
		for inx in range(self.topLevelItemCount()):
			item = self.topLevelItem(inx)
			item.setHidden(False)
		self.mark_tab_filtered(False)

	def remove_item_cb(self):
		item = self.itemFromIndex(self.last_minx)
		item.setHidden(True)
		self.mark_tab_filtered(True)

	def copy_clipboard_cb(self):
		text = ''
		for inx in range(self.topLevelItemCount()):
			item = self.topLevelItem(inx)
			if (item.isHidden() == True):
				continue
			for val in range(len(self.cmd_items)):
				text += item.column_val(val).strip() + ' '
			text = text[:-1] + '\n'

		QApplication.clipboard().setText(text)

	def get_file_line(self, minx):
		model = minx.model()
		row = minx.row()
		if (not model):
			return (None, None)
		filename = str(model.data(model.index(row, 1)))
		try:
			line = model.data(model.index(row, 2))
			line = int(line)
		except:
			line = 0
		return (filename, line)

	def add_result_continue(self):
		self.add_result(self.name, self.res)

	def add_result(self, name, res):
		res_list = []
		ret_val = True
		root = self.invisibleRootItem()
		count = 0
		for line in res:
			item = ResultPageItem(line)
			#self.addTopLevelItem(item)
			if (self.is_history):
				root.insertChild(0, item)
			else:
				root.addChild(item)
			count = count + 1
			if (count == 1000):
				self.name = name
				self.res = res[count:]
				QTimer.singleShot(100, self.add_result_continue)
				return
		self.remove_progress_bar()
		if (self.topLevelItemCount() < 5000):
			self.resizeColumnToContents(0)
			self.resizeColumnToContents(1)
			self.resizeColumnToContents(2)
			self.resizeColumnToContents(3)
		if (self.topLevelItemCount() == 1):
			item = self.topLevelItem(0)
			item.setSelected(True)
			minx = self.indexFromItem(item)
			self.activated_cb(minx)
		return ret_val

	def mousePressEvent(self, m_ev):
		QTreeWidget.mousePressEvent(self, m_ev)
		if (m_ev.button() == Qt.MouseButton.RightButton):
			self.last_minx = self.indexAt(m_ev.pos())
			self.pmenu.exec(QCursor.pos())

	def show_progress_bar(self):
		self.pbar = QProgressBar(self)
		self.pbar.setMinimum(0)
		self.pbar.setMaximum(0)
		self.pbar.show()

	def remove_progress_bar(self):
		if (self.pbar):
			self.pbar.setParent(None)
			self.pbar = None

class ResultManager(QTabWidget):
	book = None
	sig_show_file_line = pyqtSignal(str, int)

	def __init__(self, *args):
		QTabWidget.__init__(*(self, ) + args)

		self.setMovable(True)
		self.setTabsClosable(True)
		self.tabCloseRequested.connect(self.close_cb)

		self.h_page = None

		self.setFont(QFont("San Serif", 8))
		ResultManager.book = self

	def go_next_res_common(self, page, inc):
		if page == None:
			return
		if (inc == 1):
			minx = page.moveCursor(QAbstractItemView.CursorAction.MoveDown, Qt.KeyboardModifier.NoModifier)
		else:
			minx = page.moveCursor(QAbstractItemView.CursorAction.MoveUp, Qt.KeyboardModifier.NoModifier)
		page.setCurrentItem(page.itemFromIndex(minx))
		page.activated.emit(minx)
	def go_next_res(self, inc):
		self.go_next_res_common(self.currentWidget(), inc)
	def go_next_history(self, inc):
		self.go_next_res_common(self.h_page, -inc)
	def show_history(self):
		if (self.h_page):
			self.setCurrentWidget(self.h_page)

	def new_cb(self):
		print("new_cb")
	def refresh_cb(self):
		print("refresh_cb")
	def close_cb(self, inx):
		if (self.widget(inx) == self.h_page):
			self.h_page = None
		self.removeTab(inx)
	def close_all_cb(self):
		if DialogManager.show_yes_no("Close all query results ?"):
			self.clear()
			self.h_page = None

	def mousePressEvent(self, m_ev):
		QTabWidget.mousePressEvent(self, m_ev)
		if (m_ev.button() == Qt.MouseButton.RightButton):
			# setup popup menu
			self.pmenu = QMenu()
			#self.pmenu.addAction("&New", self.new_cb)
			#self.pmenu.addAction("&Refresh", self.refresh_cb)
			#self.pmenu.addSeparator()
			self.pmenu.addAction("Close &All", self.close_all_cb)
			self.pmenu.exec(QCursor.pos())

	def history_create(self):
		self.h_page = ResultManager.create_result_page(self, None)
		assert self.h_page.is_history
		self.h_page.hideColumn(0)
		self.h_page.hideColumn(3)
		
	def history_update(self, filename, line):
		if (ResultPage.is_history_call):
			return
		if (self.h_page == None):
			self.history_create()
	
		for idx in range(self.h_page.indexOfTopLevelItem(self.h_page.currentItem())):
			self.h_page.invisibleRootItem().removeChild(self.h_page.topLevelItem(0))

		if self.h_page.topLevelItemCount():
			item = self.h_page.topLevelItem(0)
			minx = self.h_page.indexFromItem(item)
			(f, l) = self.h_page.get_file_line(minx)
			if (filename and f == filename and l == line):
				return
		self.h_page.add_result('History', [['', filename, str(line), '']])
		if self.h_page.topLevelItemCount():
			self.h_page.setCurrentItem(self.h_page.topLevelItem(0))

	def create_result_page(self, name):
		page = ResultPage(self)
		if not name:
			name = 'History'
			page.is_history = True
		self.addTab(page, name)
		if not page.is_history:
			self.setCurrentWidget(page)
		page.sig_show_file_line.connect(self.sig_show_file_line)
		page.show_progress_bar()
		return page

	def create_result_page_single(self):
		return ResultPage(self)



#!/usr/bin/env python

# Copyright (c) 2014 Anil Kumar
# All rights reserved.
#
# License: BSD

import os
import backend
import CtagsManager
from collections import OrderedDict

is_debug = os.getenv('SEASCOPE_WEB_DEBUG')

g_proj_d = OrderedDict()

def p_open(proj_path):
	bl = backend.plugin_guess(proj_path)
	if len(bl) == 0:
		msg = "Project '%s': No backend is interested" % proj_path
		print(msg)
		return None
	proj_type = bl[0]
	if len(bl) > 1:
		msg = "Project '%s': Many backends interested" % proj_path
		for b in be:
			msg += '\n\t' + b.name()
		msg += '\n\nGoing ahead with: ' + proj_type
		print(msg)
	print("Project '%s': using '%s' backend" % (proj_path, proj_type))

	bp = backend.BProject()
	rc = bp.proj_open(proj_path, proj_type)
	if rc:
		return bp
	else:
		return None

def find_prj(rd):
	pdir = rd.get('project', None)
	if not pdir:
		return None
	bp = g_proj_d.get(pdir, False)
	return bp

def p_query(bp, rquery):
	return bp.proj_query(rquery)

def p_query_fl(bp):
	return bp.proj_query_fl()

def validate_op(rd):
	for r in ['cmd_type', 'cmd_str', 'req']:
		if r not in rd:
			return False
	return True

def _run_op(rd):
	if is_debug:
		print('rd', rd)

	if not validate_op:
		return { 'err_data' : 'invalid request: validate_op failed' }

	cmd_type = rd['cmd_type']

	if cmd_type == 'project_list':
		res = list(g_proj_d.keys())
		outd = { 'res' : res }
		return outd
	if cmd_type == 'query':
		bp = find_prj(rd)
		if not bp:
			return { 'err_data' : 'failed to find project for query %s' % rd }

		if rd['cmd_str'] == 'FLIST':
			outd = p_query_fl(bp)
			res = []
			res.append(['file', 'path'])
			for r in outd['res']:
				res.append([os.path.basename(r), r])
			outd['res'] = res
			return outd
		if rd['cmd_str'] == 'FDATA':
			fname = rd['req']
			pdir = bp.proj_dir()
			if not fname.startswith(pdir):
				return { 'err_data' : 'file %s not in project %s' % (fname, pdir)}
			try:
				data = open(fname).read()
				outd = { 'res' : [ data ] }
				return outd
			except:
				return { 'err_data' : 'failed to read file %s' % fname }
		if rd['cmd_str'] == 'FCTAGS':
			fname = rd['req']
			if not fname.startswith(bp.proj_dir()):
				return []
			try:
				data = CtagsManager.ct_query(rd['req'])
				outd = { 'res' : [data] }
				return outd
			except:
				pass
		rquery = {
			'cmd' : rd['cmd_str'],
			'req' : rd['req'],
			'opt' : rd.get('opt', '').split(','),
			'hint_file' : rd.get('hint_file', ''),
		}
		outd = p_query(bp, rquery)
		if is_debug:
			print('rquery', rquery)
		return outd

	return { 'err_data' : 'invalid request: unknown cmd_type=%s' % cmd_type }

def run_op(rd):
	if is_debug:
		print('-' * 80)
	outd = _run_op(rd)
	if is_debug:
		print('-' * 80)

	if ('err_data' in outd) and outd['err_data']:
		print('==> ERR_DATA:', outd['err_data'])
	res = outd.get('res', None)
	if not res and 'out_data' in outd:
		res = outd['out_data'].splitlines()
	#if not rd['cmd_str'] == 'FDATA':
		#for r in res:
			#print r
	if is_debug:
		print('-' * 80)
	return outd

def load_projects(plist):
	backend.load_plugins()

	for p in plist:
		bp = p_open(p)
		if not bp:
			return False
		global g_proj_d
		pdir = bp.proj_dir()
		g_proj_d[pdir] = bp

	if not len(g_proj_d):
		print('No projects')
		return False

	return True



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os, re, subprocess
from datetime import datetime

from PyQt6.QtCore import QThread

def _eintr_retry_call(func, *args):
	while True:
		try:
			return func(*args)
		except OSError as e:
			if e.errno == errno.EINTR:
				continue
			raise

#class CtagsInfo:
	#def __init__(self):
		#pass

	#@staticmethod
	#def lang_suffixes():
		#cmd = 'ctags --list-maps'
		#args = cmd.split()
		#import subprocess
		#proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
		#(out_data, err_data) = proc.communicate()
		#out_data = out_data.decode()
		#err_data = err_data.decode()
		#res = re.split(r'\r?\n', out_data)
		# res = out_data
		#res = [ line.split(None, 1) for line in res if line != '' ]
		#d = {}
		#for line in res:
			#import fnmatch
			#fn_pats = line[1] .split()
			#print line[0], fn_pats
			#re_pats = '|'.join(fnmatch.translate(p) for p in fn_pats)
			#print line[0], re_pats
			#d[line[0]] = re_pats
		#print d
#print 'ctag.lang'
#CtagsInfo.lang_suffixes()

class CtagsThread(QThread):
	thread_list = []

	def __init__(self, sig):
		QThread.__init__(self)
		
		self.sig = sig
		self.ct_dict = {}
		
		self.finished.connect(self._finished_cb)
		CtagsThread.thread_list.append(self)
	
	def _finished_cb(self):
		#print 'CtagsThread.finished', len(self.ct_dict)
		CtagsThread.thread_list.remove(self)
		for k in self.ct_dict:
			ct_cache[k] = self.ct_dict[k]

		self.res = self.parse_result(self.res, self.sig)
		if self.res != None:
			self.sig.emit_result(self.res)
	
	def ctags_bsearch(self, ct, n):
		x = -1
		y = len(ct) - 1

		#print f, n
		while x < y:
			m = (x + y + 1) // 2
			#print '(', x, y, m, ')'
			if ct[m][1] > n:
				#print 'y: m -1', ct[m][1]
				m = m - 1
				y = m
				continue
			if ct[m][1] < n:
				#print 'x: m', ct[m][1]
				x = m
				continue
			break
		if y == -1:
			return y
		#assert x == m or y == m or ct[m][1] == n
		return m

	def _ctags_fix(self, line):
		f = line[1]
		n = int(line[2])
		ct = self.ct_dict[f]

		m = self.ctags_bsearch(ct, n)
		if m == -1:
			return True
		if ct[m][1] != n:
			if ct[m][2] in ['variable']:
				#line[0] = '***** SMART *****'
				#print line
				return True
			if f.endswith('.py'):
				while m > 0 and (ct[m][2] in ['namespace']):
					m = m - 1
				if m == - 1:
					return True
		if ct[m][1] == n:
			if self.cmd_str == 'DEF':
				x = m
				y = m
				while x > 1 and ct[x - 1][1] == n:
					x = x - 1
				while y < len(ct) - 1 and ct[y + 1][1] == n:
					y = y + 1
				if x > y:
					return False
				for m in range(x, y + 1):
					reqPat = self.sig.sym + '$'
					if re.match(reqPat, ct[m][0]):
						line[0] = ct[m][0]
						return True
					return False
		else:
			if self.cmd_str == 'DEF':
				return False
			if self.cmd_str == '-->':
				if f.endswith('.tac'):
					if ct[m][2] in ['reactor']:
						line[0] = ct[m][0].split('=>')[0].strip() + "Is"
						return True
		line[0] = ct[m][0]
		return True

	def ct_override(self, fl):
		pattern = os.getenv('SEASCOPE_CTAGS_OVERRIDE_PATTERN')
		delim1  = os.getenv('SEASCOPE_CTAGS_OVERRIDE_DELIM_START')
		delim2  = os.getenv('SEASCOPE_CTAGS_OVERRIDE_DELIM_END')
		if not (pattern and delim1 and delim2):
			return

		res = {}
		for f in fl:
			args = ['grep', '-En', "%s" % pattern, "%s" % f]
			try:
				proc = subprocess.Popen(args, stdout=subprocess.PIPE)
				(out_data, err_data) = _eintr_retry_call(proc.communicate)
				out_data = out_data.decode()
				out_data = out_data.split('\n')
			except Exception as e:
				out_data =  [
						'Failed to run ct override cmd\tignore\t0;\t ',
						'cmd: %s\tignore\t0;\t ' % ' '.join(args),
						'error: %s\tignore\t0;\t ' % str(e),
						'ctags not installed ?\tignore\t0;\t ',
					]
			f_res = {}
			for line in out_data:
				if (line == ''):
					break
				line = line.split(':', 1)
				num = line[0]
				inx1 = line[1].find(delim1) + 1
				inx2 = line[1].find(delim2)
				sym = line[1][inx1:inx2]
				f_res[num] = [sym, num, 'function']
			res[f] = f_res
		return res

	def runCtagsCustom(self, fl):
		custom_map = os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP')
		if not custom_map:
			return []
		try:
			custom_map = eval(custom_map)
		except:
			print('SEASCOPE_CTAGS_SUFFIX_CMD_MAP has errors')
			return []

		cmd_list = []
		for (suffix, cmd) in custom_map:
			_fl = [ f for f in fl if f.endswith(suffix) ]
			args = cmd.split()
			args += _fl
			cmd_list.append(args)

		if not len(cmd_list):
			return []

		out_data_all = []
		for args in cmd_list:
			proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			(out_data, err_data) = proc.communicate('\n'.join(fl).encode())
			out_data = out_data.decode()
			out_data = re.split(r'\r?\n', out_data)
			out_data_all += out_data
		return out_data_all

	def _run_ctags(self):
		override_res = self.ct_override(self.file_list)

		cmd = 'ctags -n -u --fields=+K -L - -f -'
		opt_I_file = os.getenv('SEASCOPE_CTAGS_OPT_I_FILE')
		if opt_I_file and os.path.isfile(opt_I_file):
			cmd += ' -I ' + opt_I_file
		args = cmd.split()
		proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
		(out_data, err_data) = proc.communicate('\n'.join(self.file_list).encode())
		out_data = out_data.decode()
		out_data = re.split(r'\r?\n', out_data)
		out_data += self.runCtagsCustom(self.file_list)

		for line in out_data:
			if line == '':
				continue
			line = line.split('\t')
			f = line[1]
			num = line[2].split(';', 1)[0]
			if override_res:
				override_ent = override_res.get(f, {}).pop(num, None)
				if override_ent:
					line[0] = override_ent[0]
			line = [line[0], int(num), line[3]]
			self.ct_dict[f].append(line)
		if override_res:
			for (f, o_res) in override_res.items():
				if o_res:
					res = self.ct_dict[f]
					for line in o_res.values():
						line = [line[0], int(line[1]), line[2]]
						res.append(line)
					self.ct_dict[f] = sorted(res, key=lambda e: e[1])

	def prepare_file_list(self):
		#t1 = datetime.now()
		flist = set()
		for line in self.res:
			if line[0] in self.cond_list:
				if line[1] not in ct_cache:
					flist.add(line[1])
				else:
					self.ct_dict[line[1]] = ct_cache[line[1]]
		for f in flist:
			self.ct_dict[f] = []

		#t2 = datetime.now()
		#print '  ', len(flist), 'flist', t2 - t1

		self.file_list = flist

	def apply_fix(self, cmd_str, res, cond_list):
		self.cmd_str = cmd_str
		self.res = res
		self.cond_list = cond_list

		self.prepare_file_list()
		self.start()

	def run(self):
		#t2 = datetime.now()

		self._run_ctags()
		
		#t3 = datetime.now()
		#print '  ct', t3 - t2
		res = []
		for line in self.res:
			if line[0] not in self.cond_list:
				continue
			if not self._ctags_fix(line):
				continue
			res.append(line)
		self.res = res
		#t4 = datetime.now()
		#print '  fix', t4 - t3
		#print '  total', t4 - t2

	def _filter_res(self, res, sig):
		req = sig.sym
		out_res = []
		if self.cmd_str == 'DEF':
			import_re = re.compile(r'^\s*import\s+')
			for line in res:
				reqPat = req + '$'
				if not re.match(reqPat, line[0]):
					continue
				if import_re.search(line[3]) and line[1].endswith('.py'):
					continue
				out_res.append(line)
			return out_res
		if self.cmd_str == '-->':
			call_re = re.compile(r'\\b%s\\b\s*\(' % req)
			extern_re = re.compile(r'^\s*extern\s+')
			reactor_re = re.compile(r'\\b(\w+::)*(\w+)\s*=>.*\\b%s\\b' % req)
			comment_re = re.compile(r'^\s*(\*\s|/\*|\*/|//\s|# )')
			func_ptr_re = re.compile(r'\\b(\w+)\s*(=|:)\s*%s\s*[,;:)]' % req)
			func_as_arg_re = re.compile(r'(^\s*|[(,]\s*)(\w+(\.|->))*%s\s*[,)]' % req);
			def _check_line():
				if line[1].endswith('.tac'):
					if '=>' in line[3]:
						grp = reactor_re.search(line[3])
						if grp:
							line[0] = grp.group(2) + "Is"
							return True
					# fallthru
				if line[0] == req:
					if not re.search(r'(\.|->)%s\\b' % req, line[3]):
						return False
					return True
				if call_re.search(line[3]):
					if extern_re.search(line[3]):
						return False
					return True
				grp = func_ptr_re.search(line[3])
				if grp:
					line[0] = grp.group(1)
					return True
				if not func_as_arg_re.search(line[3]):
					False
				return True
			for line in res:
				if not _check_line():
					continue
				if line[0] == '<unknown>':
					continue
				if comment_re.search(line[3]):
					continue
				out_res.append(line)
			return out_res
		if self.cmd_str == '<--':
			return res
		if self.cmd_str == 'INC':
			inc_re = re.compile(r'^\s*(#\s*include|(from\s+[^\s]+\s+)?import)\s+.*%s.*' % req)
			for line in res:
				if not inc_re.search(line[3]):
					continue
				out_res.append(line)
			return out_res
		return res

	def parse_result(self, res, sig):
		res = self._filter_res(res, sig)
		return res

ct_cache = {}

def flush():
	#print 'flushing ctags cache...'
	global ct_cache
	ct_cache = {}




#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *
from PyQt6.QtSvgWidgets import *
import os
import sys

if __name__ == '__main__':
	import sys
	import os
	app_dir = os.path.dirname(os.path.realpath(__file__))
	os.chdir(app_dir)
	sys.path.insert(0, os.path.abspath('..'))
	os.chdir(os.path.abspath('..'))

from backend.plugins.PluginBase import PluginProcess

class FileFuncGraphProcess(PluginProcess):
	def __init__(self, wdir, rq):
		PluginProcess.__init__(self, wdir, rq)
		self.name = 'call graph process'

	def parse_result(self, text, sig):
		return [text]


class FileFuncGraphWidget(QWidget):
	def __init__(self, parent, cmd_func, cmd_id, cmd_opt):
		QWidget.__init__(self, parent)
		self.is_busy = False
		self.is_done = False
		self.cmd_func = cmd_func
		self.cmd_id = cmd_id
		self.cmd_opt = cmd_opt
		self.is_debug = os.getenv('SEASCOPE_FILE_FUNC_GRAPH_VIEW_DEBUG')

		self.vlay1 = QVBoxLayout()
		self.setLayout(self.vlay1)
		#self.hlay1 = QHBoxLayout()
		#self.vlay1.addLayout(self.hlay1)
		#self.add_buttons(self.hlay1)
		self.lbl = QLabel()
		self.vlay1.addWidget(self.lbl)

		self.vlay2 = QVBoxLayout()
		self.scrolla = QScrollArea()
		self.scrolla.setLayout(self.vlay2)
		self.vlay1.addWidget(self.scrolla)

		self.warn_experimental_lbl = QLabel('Warning: this is an EXPERIMENTAL feature based on cscope, might not work on all files for now.')
		self.warn_experimental_lbl.setStyleSheet("QLabel { color : red; }")
		self.vlay1.addWidget(self.warn_experimental_lbl)

	def startQuery(self, req, dname, proj_dir, inx):
		if self.is_done:
			return

		name = req if req else dname
		labelList = [
			'File functions graph(%s)' % name,
			'File functions and external graph(%s)' %name,
			'Directory functions graph(%s)' % os.path.dirname(name),
			'Directory functions and external graph(%s)' % os.path.dirname(name)
		]
		self.lbl.setText(labelList[inx])

		tool_path = os.path.join('tools', 'FileFuncGraph.py')
		pargs = [sys.executable, tool_path]
		if inx == 0:
			pass
		elif inx == 1:
			pargs += ['-e']
		elif inx == 2:
			dname = os.path.dirname(name)
		elif inx == 3:
			pargs += ['-e']
			dname = os.path.dirname(name)

		pargs += ['-d', dname]
		sig_res = FileFuncGraphProcess('.', None).run_query_process(pargs, req)
		sig_res[0].connect(self.clgraph_add_result)
		self.is_busy = True
		self.show_progress_bar()

	def set_current(self, btn):
		inx = self.bgrp.id(btn)
		#self.btn[inx].setChecked(True)
		#print 'inx clicked', inx
		if inx == 0:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			self.svgw.setMinimumSize(self.svgw.sizeHint())
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())
		if inx == 1:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())
		if inx == 2:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			self.svgw.resize(self.scrolla.size())
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())

	def add_buttons(self, hlay):
		self.bgrp = QButtonGroup()
		self.bgrp.buttonClicked.connect(self.set_current)
		self.bgrp.setExclusive(True)
		for inx in range(3):
			btn = QToolButton()
			btn.setText(str(inx))
			btn.setToolTip(str(inx))
			#btn.setFlat(True)
			btn.setCheckable(True)
			self.bgrp.addButton(btn, inx)
			hlay.addWidget(btn)

	def clgraph_add_result(self, req, res):
		if self.is_debug:
			print(res)
		self.is_busy = False
		self.is_done = True
		self.remove_progress_bar()
		
		self.svgw = QSvgWidget()
		self.scrolla.setWidget(self.svgw)
		self.svgw.load(QByteArray(res[0].encode()))
		
		#print self.svgw.renderer().defaultSize()
		sz = self.svgw.sizeHint()
		scale = 1
		if sz.width() > 1024:
			scale = 0.8
		self.svgw.setMinimumSize(int(sz.width() * scale), int(sz.height() * scale))
		#self.svgw.setMaximumSize(self.svgw.sizeHint())
		#print self.scrolla.sizeHint()

	def show_progress_bar(self):
		self.pbar = QProgressBar(self.scrolla)
		self.pbar.setMinimum(0)
		self.pbar.setMaximum(0)
		self.pbar.show()

	def remove_progress_bar(self):
		if self.pbar:
			self.pbar.hide()
			self.pbar.setParent(None)
			self.pbar = None

class FileFuncGraphWindow(QMainWindow):
	parent = None

	def __init__(self, req, dname, proj_dir, cmd_func, cmd_args, cmd_opt):
		QMainWindow.__init__(self, FileFuncGraphWindow.parent)
		self.req = req
		self.dname = dname
		self.proj_dir = proj_dir

		if req:
			self.setWindowTitle(req)
		else:
			self.setWindowTitle(dname)

		self.setFont(QFont("San Serif", 8))

		w = QWidget()
		self.setCentralWidget(w)
		self.vlay = QVBoxLayout()
		w.setLayout(self.vlay)

		self.sw = QStackedWidget()

		self.hlay = QHBoxLayout()
		self.vlay.addLayout(self.hlay)
		self.vlay.addWidget(self.sw)
		
		self.bgrp = QButtonGroup()
		self.bgrp.buttonClicked.connect(self.set_current)
		self.bgrp.setExclusive(True)

		self.btn = []
		self.ctree = []
		for inx in range(len(cmd_args)):
			# cmd format: [ cmd_id, cmd_str, cmd_tip ]
			cmd = cmd_args[inx]

			btn = QToolButton()
			btn.setText(cmd[1])
			btn.setToolTip(cmd[2])
			#btn.setFlat(True)
			btn.setCheckable(True)
			self.bgrp.addButton(btn, inx)
			self.hlay.addWidget(btn)

			ct = FileFuncGraphWidget(self, cmd_func, cmd[0], cmd_opt)
			self.sw.addWidget(ct)

			self.btn.append(btn)
			self.ctree.append(ct)
		self.hlay.addStretch(0)
		self.set_current(self.btn[0])

	def set_current(self, btn):
		inx = self.bgrp.id(btn)
		self.btn[inx].setChecked(True)
		self.sw.setCurrentIndex(inx)
		ct = self.ctree[inx]
		ct.setFocus()
		
		ct.startQuery(self.req, self.dname, self.proj_dir, inx)



def create_page(req, dname, proj_dir, cmd_func, cmd_args, cmd_opt):
	w = FileFuncGraphWindow(req, dname, proj_dir, cmd_func, cmd_args, cmd_opt)
	w.resize(900, 600)
	w.show()
	return w

if __name__ == '__main__':
	import optparse
	usage = "usage: %prog (-d <code_dir/file> | -p <idutils_proj>) [symbol]"
	op = optparse.OptionParser(usage=usage)
	op.add_option("-d", "--codedir", dest="code_dir", help="Code dir", metavar="CODE_DIR")
	op.add_option("-p", "--project", dest="id_path", help="Idutils project dir", metavar="PROJECT")
	(options, args) = op.parse_args()

	sym = ''
	dname = ''
	id_path = None

	if (not any([options.code_dir, options.id_path]) or
		all([options.code_dir, options.id_path])):
		print('Specify one among -d or -p', file=sys.stderr)
		sys.exit(-1)

	if len(args):
		if len(args) != 1:
			print('Please specify a symbol', file=sys.stderr)
			sys.exit(-4)
		sym = args[0]

	if options.code_dir:
		dname = options.code_dir
		if not os.path.exists(dname):
			print('"%s": does not exist' %  dname, file=sys.stderr)
			sys.exit(-2)
	if options.id_path:
		if not sym:
			print('-p option needs a symbol', file=sys.stderr)
			sys.exit(-3)
		id_path = os.path.normpath(options.id_path)
		if not os.path.exists(os.path.join(id_path, 'ID')):
			print('idutils project path does not exist', file=sys.stderr)
			sys.exit(-4)

	app = QApplication(sys.argv)
	cmd_args = [
		['FFGRAPH',    'F',   'File functions graph'],
		['FFGRAPH_E',  'F+E', 'File functions + external graph'],
		['FFGRAPH_D',  'D',   'Directory functions graph'],
		['FFGRAPH_DE', 'D+E', 'Directory functions + external graph']
	]

	w = create_page(sym, dname, id_path, None, cmd_args, None)

	sys.exit(app.exec())



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import textwrap
from distutils.core import setup

setup(name='Seascope',
	version='0.9',
	description='A multi-platform multi-language source code browsing tool',
	long_description= textwrap.dedent("""A pyQt GUI front-end for idutils, cscope and gtags. Written in python using pyQt, QScintilla libraries."""),
	url='https://github.com/anilFree/seascope',
	packages=['Seascope',
		'Seascope.backend',
		'Seascope.backend.plugins',
		'Seascope.backend.plugins.idutils',
		'Seascope.backend.plugins.cscope',
		'Seascope.backend.plugins.gtags',
		'Seascope.view',
                'Seascope.view.filecontext',
                'Seascope.view.filecontext.plugins',
                'Seascope.view.filecontext.plugins.ctags_view',
                'Seascope.view.filecontext.plugins.generic_view',
		],
	package_dir={'Seascope': 'src'},
	package_data={'Seascope': ['icons/*.svg','tools/*.py', 'ui/*.ui', 'backend/plugins/*/ui/*.ui']},
	license="BSD License",
	classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: X11 Applications :: Qt',
          'Environment :: Win32 (MS Windows)',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Software Development',
          ],
	options= {'bdist_rpm':{'requires': 'PyQt5,python3-pyqt5.qsci,idutils,cscope,global,ctags,graphviz',
				'group': 'Development Tools',
				'vendor': 'The Seascope Team'}}
	)



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import sys
sys.dont_write_bytecode = True

import os
import string

try:
	from PyQt6 import QtWidgets, QtGui, QtCore
except ImportError:
	print('Error: PyQt6 package not found\nError: required packages: PyQt6 and python3-pyqt6.qsci\nError: program aborted.')
	sys.exit(-1)

try:
	from PyQt6.QtWidgets import *
	from PyQt6.QtGui import *
	from PyQt6.QtCore import *
	from PyQt6 import uic
	from view import EdView, EdViewRW, ResView, FileView, CallView, ClassGraphView, FileFuncGraphView
	from view import DebugView, CodemarkView, CodeContextView
	from view import ProjectUi
	import backend
	import DialogManager
	import view
except ImportError as e:
	print(str(e))
	print("Error: failed to import supporting packages.\nError: program aborted.")
	sys.exit(-1)

def msg_box(msg):
	QMessageBox.warning(None, "Seascope", msg, QMessageBox.StandardButton.Ok)

class BackendChooserDialog(QDialog):
	def __init__(self):
		QDialog.__init__(self)
		self.ui = uic.loadUi('ui/proj_new.ui', self)
		self.backend_lw.currentRowChanged.connect(self.currentRowChanged_cb)

		self.backend = backend

	def currentRowChanged_cb(self, row):
		if row == -1:
			return
		bname = str(self.backend_lw.currentItem().text())
		desc = ''
		for b in self.backend.plugin_list():
			if b.name() == bname:
				desc = b.description()
		self.descr_te.setText(desc)

	def run_dialog(self):
		blist = self.backend.plugin_list()
		if len(blist) == 0:
			msg_box('No backends are available/usable')
			return None
		bi = [ b.name() for b in blist ]
		self.backend_lw.addItems(bi)
		self.backend_lw.setCurrentRow(0)
		if self.exec() == QDialog.DialogCode.Accepted:
			bname = str(self.backend_lw.currentItem().text())
			return bname
		return None

class QueryUiHelper:
	def __init__(self):
		self.edit_book = None
		self.res_book = None
		self.call_view = None
		self.class_graph_view = None
		self.file_func_graph_view = None
		self.dbg_view = None
		self.file_view = None

	def editor_current_file(self):
		(f, l) = self.edit_book.get_current_file_line()
		return f

	def editor_current_word(self):
		return self.edit_book.get_current_word()

	def result_page_new(self, name, sig_res):
		if sig_res == None:
			return
		page = self.res_book.create_result_page(name)
		## add result
		sig_res[0].connect(page.add_result)
		#page.add_result(req, res)
		self.dbg_view.connect_to_sig(sig_res[1])

	def file_view_update(self, flist):
		self.file_view.add_files(flist)

	def _quick_def_result(self, req, res):
		count = len(res)
		if count > 1:
			page = self.res_book.create_result_page_single()
			page.add_result(req, res)
			
			dlg = QDialog()
			dlg.setWindowTitle('Quick Definition: ' + req)
			vlay = QVBoxLayout(dlg)
			vlay.addWidget(page)

			page.sig_show_file_line.connect(self.edit_book.show_file_line)
			page.activated.connect(dlg.accept)
			page.setMinimumWidth(800)
			page.setMinimumHeight(100)

			dlg.exec()
			return

		if (count == 1):
			filename = res[0][1]
			try:
				line = int(res[0][2])
			except:
				return
			self.edit_book.show_file_line(filename, line)

	def quick_def_page_new(self, sig_res):
		if sig_res == None:
			return
		sig_res[0].connect(self._quick_def_result)
		self.dbg_view.connect_to_sig(sig_res[1])

	def call_view_page_new(self, req, query_func, ctree_query_args, opt):
		hint_file = self.editor_current_file()
		self.call_view.create_page(req, query_func, ctree_query_args, opt, hint_file)

	def class_graph_view_page_new(self, req, dname, prj_type, query_func, clgraph_query_args, opt):
		self.class_graph_view.create_page(req, dname, prj_type, query_func, clgraph_query_args, opt)

	def file_func_graph_view_page_new(self, req, dname, proj_dir, query_func, ffgraph_query_args, opt):
		self.file_func_graph_view.create_page(req, dname, proj_dir, query_func, ffgraph_query_args, opt)


class QueryDialogUi(QDialog):
	def __init__(self, title, feat):
		QDialog.__init__(self)

		self.feat = feat
		self.ui = uic.loadUi('ui/query.ui', self)
		self.qd_sym_inp.setCompleter(None)
		self.qd_sym_inp.setInsertPolicy(QComboBox.InsertPolicy.InsertAtTop)

		self.setWindowTitle(title)
		self.qd_cmd_inp.addItems(self.feat.cmd_qstrlist)

	def run_dialog(self, cmd_str, req):
		qstr = self.feat.cmd_str2qstr[cmd_str]
		inx = self.qd_cmd_inp.findText(qstr)
		self.qd_cmd_inp.setCurrentIndex(inx)
		if req == None:
			req = ''
		self.qd_sym_inp.setFocus()
		self.qd_sym_inp.setEditText(req)
		self.qd_sym_inp.lineEdit().selectAll()
		self.qd_substr_chkbox.setChecked(False)
		self.qd_icase_chkbox.setChecked(False)

		self.show()
		if self.exec() == QDialog.DialogCode.Accepted:
			req = str(self.qd_sym_inp.currentText())
			cmd = str(self.qd_cmd_inp.currentText())
			cmd_str = self.feat.cmd_qstr2str[cmd]
			#self.qd_sym_inp.addItem(req)
			in_opt = {
				'substring'   : self.qd_substr_chkbox.isChecked(),
				'ignorecase'  : self.qd_icase_chkbox.isChecked(),
			}
			res = self.feat.query_dlg_cb(req, cmd_str, in_opt)
			return res
		return None

	def show_dlg(self, cmd_str, req):
		return self.run_dialog(cmd_str, req)

class QueryUi(QObject):
	def __init__(self, h):
		QObject.__init__(self)
		self.qui_h = h
		self.backend = backend
		self.backend_menu = None
		self.prj_actions = []

		self.feat = None
		self.qry_ui = None

	def set_backend_menu(self, menu):
		self.backend_menu = menu
		self.backend_menu.triggered.connect(self.menu_cb)

	def setup(self):
		self.prepare_menu()
		for act in self.prj_actions:
			act.setEnabled(True)

		self.feat = self.backend.proj_feature()
		t = backend.proj_type().title() + ' Query'
		self.qry_ui = QueryDialogUi(t, self.feat)

		self.query_file_list()

	def reset(self):
		self.backend_menu.clear()
		self.backend_menu.setTitle('')
		for act in self.prj_actions:
			act.setEnabled(False)
		
		self.feat = None
		self.qry_ui = None

	def editor_current_file(self):
		return self.qui_h.editor_current_file()

	def editor_current_word(self):
		return self.qui_h.editor_current_word()

	def prepare_menu(self):
		td = {
			'idutils' : '&Idutils',
			'gtags'   : 'G&tags',
			'cscope'  : '&Cscope',
		}
		t = td[self.backend.proj_type()]
		menu = self.backend_menu
		menu.setTitle(t)
		feat = self.backend.proj_feature()
		for c in feat.menu_cmd_list:
			if c[0] == '---':
				menu.addSeparator()
				continue
			if c[2] == None:
				if c[0] == 'UPD':
					func = self.rebuild_cb
					act = menu.addAction(c[1], func)
				act.cmd_str = None
			else:
				act = menu.addAction(c[1])
				act.setShortcut(c[2])
				act.cmd_str = c[0]

	def menu_cb(self, act):
		if act.cmd_str != None:
			self.query_cb(act.cmd_str)

	def query_cb(self, cmd_str):
		if not backend.proj_conf():
			return
		if not backend.proj_is_ready():
			show_msg_dialog('\nProject has no source files')
			return

		if cmd_str == 'CLGRAPHD':
			f = self.editor_current_file()
			if f:
				d = os.path.dirname(f)
				self.query_class_graph_dir(d)
			return
		if cmd_str == 'FFGRAPH':
			f = self.editor_current_file()
			if f:
				self.query_file_func_graph(f)
			return

		req = self.editor_current_word()
		if (req != None):
			req = str(req).strip()
		opt = None
		if cmd_str not in [ 'QDEF' ]:
			val = self.qry_ui.show_dlg(cmd_str, req)
			if val == None:
				return
			(cmd_str, req, opt) = val
		if (req == None or req == ''):
			return

		if cmd_str == 'QDEF':
			self.query_qdef(req, opt)
		elif cmd_str == 'CTREE':
			self.query_ctree(req, opt)
		elif cmd_str == 'CLGRAPH':
			self.query_class_graph(req, opt)
		else:
			self.do_query(cmd_str, req, opt)

	def do_proj_query(self, rquery):
		print('do_proj_query', rquery)
		sig_res = self.backend.proj_query(rquery)
		return sig_res

	def query_ctree(self, req, opt):
		self.qui_h.call_view_page_new(req, self.do_proj_query, self.feat.ctree_query_args, opt)

	def query_class_graph(self, req, opt):
		prj_type = self.backend.proj_type()
		prj_dir = self.backend.proj_dir()
		self.qui_h.class_graph_view_page_new(req, prj_dir, prj_type, self.do_proj_query, self.feat.clgraph_query_args, opt)

	def query_class_graph_dir(self, dname):
		opt = []
		self.qui_h.class_graph_view_page_new('', dname, None, self.do_proj_query, self.feat.clgraph_query_args, opt)

	def query_file_func_graph(self, fname):
		opt = []
		self.qui_h.file_func_graph_view_page_new('', fname, '', self.do_proj_query, self.feat.ffgraph_query_args, opt)

	def _prepare_rquery(self, cmd_str, req, opt):
		rquery = {}
		rquery['cmd'] = cmd_str
		rquery['req'] = req
		rquery['opt'] = opt
		# add current file info
		rquery['hint_file'] = self.editor_current_file()
		return rquery

	def query_qdef(self, req, opt):
		rquery = {}
		rquery = self._prepare_rquery('DEF', req, opt)
		sig_res = self.do_proj_query(rquery)
		self.qui_h.quick_def_page_new(sig_res)

	def do_query(self, cmd_str, req, opt):
		## create page
		name = cmd_str + ' ' + req
		rquery = self._prepare_rquery(cmd_str, req, opt)
		sig_res = self.do_proj_query(rquery)
		self.qui_h.result_page_new(name, sig_res)

	def do_rebuild(self):
		sig_rebuild = self.backend.proj_rebuild()
		if not sig_rebuild:
			return
		dlg = QProgressDialog()
		dlg.setWindowTitle('Seascope rebuild')
		dlg.setLabelText('Rebuilding database...')
		dlg.setCancelButton(None)
		dlg.setMinimum(0)
		dlg.setMaximum(0)
		sig_rebuild.connect(dlg.accept)
		while dlg.exec() != QDialog.DialogCode.Accepted:
			pass
		self.query_file_list()


	def rebuild_cb(self):
		self.do_rebuild()

	def query_file_list(self):
		sig_res = self.backend.proj_query_fl()
		if sig_res == None:
			return
		if isinstance(sig_res , list):
			fl = sig_res
			self.qui_h.file_view_update(fl)
			return
		sig_res.connect(self.qui_h.file_view_update)

class SeascopeApp(QMainWindow):

	def file_preferences_cb(self):
		ev_font = QFont()
		ev_font.fromString(self.edit_book.ev_font)
		res = DialogManager.show_preferences_dialog(self.app_style, self.edit_ext_cmd, ev_font, self.exit_dont_ask, self.inner_editing, self.eb_is_show_line)
		(self.app_style, self.app_font, self.edit_ext_cmd, self.ev_font, self.exit_dont_ask, self.inner_editing_conf, self.eb_is_show_line) = res
		if self.edit_ext_cmd != None:
			self.edit_ext_cmd = str(self.edit_ext_cmd).strip()
		if (self.edit_ext_cmd == None or self.edit_ext_cmd == ''):
			self.edit_ext_cmd = 'x-terminal-emulator -e vim %F +%L'
		self.edit_book.change_ev_font(self.ev_font)
		self.code_ctx_view.change_ev_font(self.ev_font)
		self.edit_book.show_line_number_pref(self.eb_is_show_line)
		self.app_write_config()

	def file_close_cb(self):
		self.edit_book.close_current_page()
	def closeEvent(self, ev):
		if not self.exit_dont_ask and backend.proj_is_open():
			ret = DialogManager.show_yes_no_dontask('Close project and quit?')
			if ret == 1:
				ev.ignore()
				return
			if ret == 2:
				self.exit_dont_ask = True

		# extra proc for editing enabled
		if self.inner_editing:
			self.edit_book.close_all_cb()

		self.app_write_config()
		if backend.proj_is_open():
			self.app_gui_state_save(backend.proj_dir())
		ev.accept()
	def file_restart_cb(self):
		if not DialogManager.show_yes_no('Restart ?'):
			return
		hint = self.edit_book.get_file_line_list()
		if backend.proj_is_open():
			self.proj_close_cb()
		QApplication.quit()
		os.environ['SEASCOPE_RESTART_HINT'] = '%s' % str(hint)
		QProcess.startDetached(sys.executable, QApplication.arguments(), self.seascope_start_dir);
	def file_restarted_cb(self):
		try:
			hint = os.environ['SEASCOPE_RESTART_HINT']
			#del os.environ['SEASCOPE_RESTART_HINT']
			fll = eval(hint)
			for fl in fll:
				try:
					(f, l) = fl.rsplit(':', 1)
					l = int(l)
					self.edit_book.show_file_line(f, l, hist=False)
				except Exception as e:
					print(e)
		except:
			pass

	def codemark_add(self, f, l):
		self.cm_mgr.append(f, l)
		self.edit_book.codemark_add(f, l - 1)
		actionText = f + " : " + str(l)
		act = QAction(actionText, self)
		self.cm_actionGroup.addAction(act)
		self.m_cm.addAction(act)

	def codemark_delete(self, f, l):
		self.cm_mgr.delete(f, l)
		self.edit_book.codemark_del(f, l - 1)
		actionText = f + " : " + str(l)
		actions = self.cm_actionGroup.actions()
		for act in actions:
			if actionText == act.text():
				self.cm_actionGroup.removeAction(act)
		
	def codemark_toggle_cb(self):
		(f, l) = self.edit_book.get_current_file_line()
		if self.cm_mgr.check(f, l) == 0:
			self.codemark_add(f, l)
		else:
			self.codemark_delete(f, l)

	def codemark_del_file_cb(self, filename):
		if filename == '':
			return
		for f, l in reversed(self.cm_mgr.codemarks()):
			if f == filename:
				self.codemark_delete(f, l)

	def codemark_del_all_cb(self):
		if not self.cm_mgr.count():
			return
		if not DialogManager.show_yes_no('Delete all codemarks ?'):
			return
		for f, l in reversed(self.cm_mgr.codemarks()):
			self.codemark_delete(f, l)
		
	def codemark_go(self, action):
		for f, l in self.cm_mgr.codemarks():
			actionText = f + " : " + str(l)
			if action.text() == actionText:
				self.edit_book.show_file_line(f, l)
	def go_prev_res_cb(self):
		self.res_book.go_next_res(-1)
	def go_next_res_cb(self):
		self.res_book.go_next_res(+1)
	def go_prev_pos_cb(self):
		self.res_book.go_next_history(-1)
	def go_next_pos_cb(self):
		self.res_book.go_next_history(+1)
	def go_pos_history_cb(self):
		self.res_book.show_history()
	def go_search_file_list_cb(self):
		self.file_view.search_file_cb()
	def go_search_ctags_cb(self):
		self.edit_book.focus_search_ctags()

	def help_about_cb(self):
		DialogManager.show_about_dialog()

	def external_editor_cb(self):
		self.edit_book.open_in_external_editor(self.edit_ext_cmd)

	def show_dbg_dialog(self):
		DebugView.show_dbg_dialog(self)

	def create_mbar(self):
		menubar = self.menuBar()

		m_file = menubar.addMenu('&File')
		m_file.addAction('&Preferences', self.file_preferences_cb)
		m_file.addAction('&Debug', 'Ctrl+D', self.show_dbg_dialog)
		m_file.addSeparator()
		if self.inner_editing:
			m_file.addAction('&Save', self.edit_book.save_current_page, 'Ctrl+S')
		m_file.addAction('&Close', 'Ctrl+W', self.file_close_cb)
		m_file.addSeparator()
		m_file.addAction('&Restart', 'Ctrl+R', self.file_restart_cb)
		m_file.addAction('&Quit', 'Ctrl+Q', self.close)

		m_edit = menubar.addMenu('&Edit')
		
		if self.inner_editing:
			m_edit.addAction('Undo', 'Ctrl+Z', self.edit_book.undo_edit_cb)
			m_edit.addAction('Rebo', 'Ctrl+Y', self.edit_book.redo_edit_cb)
			m_edit.addSeparator()
		m_edit.addAction('Copy', 'Ctrl+C', self.edit_book.copy_edit_cb)
		if self.inner_editing:
			m_edit.addAction('Paste', 'Ctrl+V', self.edit_book.paste_edit_cb)
			m_edit.addAction('Cut', 'Ctrl+X', self.edit_book.cut_edit_cb)
		m_edit.addSeparator()	

		m_edit.addAction('&Find...', 'Ctrl+F', self.edit_book.find_cb)
		m_edit.addAction('Find &Next', 'F3', self.edit_book.find_next_cb)
		m_edit.addAction('Find &Previous', 'Shift+F3', self.edit_book.find_prev_cb)

		m_edit.addSeparator()
		self.edit_book.m_show_line_num = m_edit.addAction('Show line number', 'F11', self.edit_book.show_line_number_cb)
		self.edit_book.m_show_line_num.setCheckable(True)
		self.edit_book.m_show_line_num.setChecked(True)
		self.show_toolbar = m_edit.addAction('Show toolbar', 'F4', self.show_toolbar_cb)
		self.show_toolbar.setCheckable(True)

		m_edit.addSeparator()
		self.edit_book.m_show_folds = m_edit.addAction('Show folds', self.edit_book.show_folds_cb)
		self.edit_book.m_show_folds.setCheckable(True)
		self.toggle_folds = m_edit.addAction('Toggle folds', self.edit_book.toggle_folds_cb)
		
		m_edit.addSeparator()
		m_edit.addAction('Refresh file', 'F5', self.edit_book.refresh_file_cb)
		m_edit.addAction('Matching brace', 'Ctrl+6', self.edit_book.matching_brace_cb)
		m_edit.addAction('Goto line', 'Ctrl+G', self.edit_book.goto_line_cb)
		m_edit.addAction('External editor', 'Ctrl+E', self.external_editor_cb);

		m_prj = menubar.addMenu('&Project')
		m_prj.addAction('&New Project', self.proj_new_cb)
		m_prj.addAction('&Open Project', self.proj_open_cb)
		m_prj.addSeparator()
		
		act = m_prj.addAction('&Settings', self.proj_settings_cb)
		act.setDisabled(True)
		self.qui.prj_actions.append(act)
		act = m_prj.addAction('&Close Project', self.proj_close_cb)
		act.setDisabled(True)
		self.qui.prj_actions.append(act)

		self.backend_menu = menubar.addMenu('')

		self.m_cm = menubar.addMenu('&Codemark')
		self.m_cm.addAction('Toggle codemark', 'Ctrl+B', self.codemark_toggle_cb)
		self.m_cm.addAction('Delete all codemarks', '', self.codemark_del_all_cb)
		self.m_cm.addSeparator()
		self.m_cm.addAction('Previous bookmark', 'Alt+PgUp', self.edit_book.bookmark_prev_cb)
		self.m_cm.addAction('Next bookmark', 'Alt+PgDown', self.edit_book.bookmark_next_cb)
		self.m_cm.addSeparator()
		self.cm_actionGroup = QActionGroup(self, triggered=self.codemark_go)

		m_go = menubar.addMenu('&Go')
		m_go.addAction('Previous Result', 'Alt+Up', self.go_prev_res_cb)
		m_go.addAction('Next Result', 'Alt+Down', self.go_next_res_cb)
		m_go.addSeparator()
		m_go.addAction('Previous Position', 'Alt+Left', self.go_prev_pos_cb)
		m_go.addAction('Next Position', 'Alt+Right', self.go_next_pos_cb)
		m_go.addAction('Position History', 'Ctrl+H', self.go_pos_history_cb)
		m_go.addSeparator()
		m_go.addAction('Search file list', 'Ctrl+Shift+O', self.go_search_file_list_cb)
		m_go.addAction('Search ctags', 'Ctrl+Shift+T', self.go_search_ctags_cb)
		
		m_help = menubar.addMenu('&Help')
		m_help.addAction('About Seascope', self.help_about_cb)
		m_help.addAction('About Qt', QApplication.aboutQt)

	def create_toolbar(self):
		self.toolbar = self.addToolBar('Toolbar')
		self.toolbar.setIconSize(QSize(16,16))
		self.toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

		si = QApplication.style().standardIcon
		#(QStyle.StandardPixmap.SP_DirClosedIcon)

		# exit
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_TitleBarCloseButton), 'Quit', self.close)

		# edit related
		# if need editing support
		if self.inner_editing:
			self.toolbar.addSeparator()
			self.toolbar.addAction(si(QStyle.StandardPixmap.SP_DialogSaveButton), 'Save', self.edit_book.save_current_page)
			self.toolbar.addSeparator()
			self.toolbar.addAction(QIcon('icons/undo.png'), 'Undo', self.edit_book.undo_edit_cb)
			self.toolbar.addAction(QIcon('icons/redo.png'), 'Redo', self.edit_book.redo_edit_cb)
			self.toolbar.addSeparator()
			self.toolbar.addAction(QIcon('icons/cut.png'), 'Cut', self.edit_book.cut_edit_cb)
			self.toolbar.addAction(QIcon('icons/copy.png'), 'Copy', self.edit_book.copy_edit_cb)
			self.toolbar.addAction(QIcon('icons/paste.png'), 'Paste', self.edit_book.paste_edit_cb)
			self.toolbar.addSeparator()
			self.toolbar.addAction(QIcon('icons/find-replace.png'), 'Find & Replace', self.edit_book.find_cb)

		# find 
		self.toolbar.addSeparator()
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_FileDialogContentsView), 'Find', self.edit_book.find_cb)
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_ArrowDown), 'Find Next', self.edit_book.find_next_cb)
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_ArrowUp), 'Find Previous', self.edit_book.find_prev_cb)
		self.toolbar.addSeparator()
		# goto
		self.toolbar.addAction(QIcon('icons/go-jump.png'), 'Go to line', self.edit_book.goto_line_cb)
		self.toolbar.addSeparator()

		# code view
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_FileDialogListView), 'Search ctags', self.go_search_ctags_cb)
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_FileDialogDetailedView), 'Search file list', self.go_search_file_list_cb)
		self.toolbar.addSeparator()
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_MediaSeekBackward), 'Previous Result', self.go_prev_res_cb)
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_MediaSeekForward), 'Next Result', self.go_next_res_cb)
		self.toolbar.addSeparator()
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_MediaSkipBackward), 'Previous Position', self.go_prev_pos_cb)
		self.toolbar.addAction(si(QStyle.StandardPixmap.SP_MediaSkipForward), 'Next Position', self.go_next_pos_cb)
		self.toolbar.addSeparator()
		self.code_ctx_view_act = self.toolbar.addAction(QIcon('icons/codeview.png'), 'Code Quick View', self.code_ctx_view_act_cb)
		self.code_ctx_view_act.setCheckable(True)
		self.code_ctx_view_act.setChecked(self.is_show_code_ctx_view)
	
		
	# app config
	def app_get_config_file(self):
		config_file = '~/.seascoperc'
		return os.path.expanduser(config_file)

	def app_read_config(self):
		self.recent_projects = []
		self.app_style = None
		self.app_font = None
		self.ev_font = None
		self.exit_dont_ask = False
		self.inner_editing_conf = False
		self.inner_editing = False
		self.is_show_toolbar = False
		self.is_show_code_ctx_view = False
		self.edit_ext_cmd = 'x-terminal-emulator -e vim %F +%L'
		self.eb_is_show_line = False

		path = self.app_get_config_file()
		if (not os.path.exists(path)):
			return
		cf = open(path, 'r')
		for line in cf:
			line = line.split('=', 1)
			key = line[0].strip()
			if (key == 'recent_projects'):
				for path in line[1].split(','):
					path=path.strip()
					if (os.path.isdir(path)):
						self.recent_projects.append(path)
			if (key == 'app_style'):
				self.app_style = line[1].split('\n')[0]
			if (key == 'app_font'):
				self.app_font = line[1].split('\n')[0]
			if (key == 'edit_font'):
				self.ev_font = line[1].split('\n')[0]
			if (key == 'edit_show_line_num'):
				if ('true' == line[1].split('\n')[0]):
					self.eb_is_show_line = True
			if (key == 'show_toolbar'):
				if ('true' == line[1].split('\n')[0]):
					self.is_show_toolbar = True
				else:
					self.is_show_toolbar = False
			if (key == 'edit_ext_cmd'):
				self.edit_ext_cmd = line[1]
			if (key == 'exit_dont_ask'):
				if ('true' == line[1].split('\n')[0]):
					self.exit_dont_ask = True
			if (key == 'inner_editing'):
				if ('true' == line[1].split('\n')[0]):
					self.inner_editing_conf = True
				else:
					self.inner_editing_conf = False
			if os.getenv("SEASCOPE_EDIT"):
				self.inner_editing = True
			else:
				self.inner_editing = self.inner_editing_conf
		cf.close()

	def app_write_config(self):
		cf = open(self.app_get_config_file(), 'w')
		cf.write('recent_projects' + '=' + ','.join(self.recent_projects)+ '\n')
		if (self.app_style):
			cf.write('app_style' + '=' + self.app_style + '\n')
		if (self.app_font):
			cf.write('app_font' + '=' + self.app_font + '\n')
		if (self.ev_font):
			cf.write('edit_font' + '=' + self.ev_font + '\n')
		if (self.inner_editing_conf):
			cf.write('inner_editing' + '=' + 'true' + '\n')
		else:
			cf.write('inner_editing' + '=' + 'false' + '\n')
		if (self.eb_is_show_line):
			cf.write('edit_show_line_num' + '=' + 'true' + '\n')
		if (self.is_show_toolbar):
			cf.write('show_toolbar' + '=' + 'true' + '\n')
		if (self.exit_dont_ask):
			cf.write('exit_dont_ask' + '=' + 'true' + '\n')
		if (self.edit_ext_cmd):
			cf.write('edit_ext_cmd' + '=' + self.edit_ext_cmd + '\n')
		cf.close()

	def app_gui_state_file(self):
		return os.path.expanduser('~/.seascopeguirc')

	def app_gui_state_read(self):
		f = self.app_gui_state_file()
		data = None
		try:
			if os.path.exists(f):
				with open(f, 'rb') as fp:
					import json
					data = json.load(fp)
		except Exception as e:
			print('app_gui_state_read:', e)
		return data

	def app_gui_state_write(self, data):
		f = self.app_gui_state_file()
		try:
			import io
			import json
			sp = io.StringIO()
			json.dump(data, sp)
			with open(f, 'wb') as fp:
				fp.write(sp.getvalue().encode())
		except Exception as e:
			print('app_gui_state_write:', e)

	def app_gui_get_geometry(self):
		rect = self.geometry()
		return (rect.x(), rect.y(), rect.width(), rect.height())
	
	def app_gui_set_geometry(self, geometry):
		x = geometry[0]
		y = geometry[1]
		w = geometry[2]
		h = geometry[3]
		self.setGeometry(x, y, w, h)

	def app_gui_state_save(self, proj_path):
		pd                    = {}
		pd ['file_line_list'] = self.edit_book.get_file_line_list()
		pd ['dir_view_list']  = self.file_view.get_dir_view_list()
		pd ['geometry']       = self.app_gui_get_geometry()

		try:
			data = self.app_gui_state_read()
			if data == None:
				data = {}
			if 'proj_gui_state' not in data:
				data['proj_gui_state'] = {}
			data['proj_gui_state'][proj_path] = pd
			self.app_gui_state_write(data)
		except Exception as e:
			print('app_gui_state_save:', e)

	def app_gui_state_restore(self, proj_path):
		try:
			data = self.app_gui_state_read()
			if data == None:
				return
			pgs = data.pop('proj_gui_state', {})
			pd = pgs.pop(proj_path, {})
			for fl in pd.pop('file_line_list', []):
				try:
					(f, l) = fl.rsplit(':', 1)
					l = int(l)
					if os.path.isabs(f) and f.startswith(proj_path) and os.path.isfile(f):
						self.edit_book.show_file_line(f, l, hist=False)
				except:
					pass
			for d in pd.pop('dir_view_list', []):
				if os.path.isabs(d) and f.startswith(proj_path) and os.path.isdir(d):
					self.file_view.open_dir_view(d)
			try:
				self.app_gui_set_geometry(pd ['geometry'])
			except Exception as e:
				print('Geometry configuration not restored')
				
		except Exception as e:
			print('app_gui_state_restore:', e)
		
	def update_recent_projects(self, path):
		if (path == None or path == ""):
			return
		new_list = [path]
		for ele in self.recent_projects:
			if ele != path:
				new_list.append(ele)
		self.recent_projects = new_list
		self.app_write_config()


	# project menu functions
	def proj_new_or_open(self):
		self.qui.setup()

		self.editor_tab_changed_cb(None)
		proj_path = backend.proj_dir()
		self.update_recent_projects(proj_path)
		self.app_gui_state_restore(proj_path)

	def proj_new_cb(self):
		if backend.proj_is_open():
			if not DialogManager.show_proj_close():
				return
			self.proj_close_cb()

		dlg = BackendChooserDialog()
		bname = dlg.run_dialog()
		if bname == None:
			return

		proj_args = ProjectUi.show_settings_ui(bname, None)
		if not proj_args:
			return

		if not backend.proj_new(bname, proj_args):
			return
		# XXX FIXIT
		self.qui.do_rebuild()
		self.proj_new_or_open()

	def proj_open(self, proj_path):
		bl = backend.plugin_guess(proj_path)
		if len(bl) == 0:
			msg = "Project '%s': No backend is interested" % proj_path
			msg_box(msg)
			return
		proj_type = bl[0]
		if len(bl) > 1:
			msg = "Project '%s': Many backends interested" % proj_path
			for b in bl:
				msg += '\n\t' + b
			msg += '\n\nGoing ahead with: ' + proj_type
			msg_box(msg)
		print("Project '%s': using '%s' backend" % (proj_path, proj_type))

		rc = backend.proj_open(proj_path, proj_type)
		if not rc:
			print('proj_open', proj_path, 'failed')
			return

		self.proj_new_or_open()

	def proj_open_cb(self):
		proj_path = DialogManager.show_project_open_dialog(self.recent_projects)
		if proj_path != None:
			if backend.proj_is_open():
				self.proj_close_cb()

			self.proj_open(proj_path)

	def proj_close_cb(self):
		proj_path = backend.proj_dir()
		self.update_recent_projects(proj_path)
		self.app_gui_state_save(proj_path)
		self.setWindowTitle("Seascope")

		backend.proj_close()
		self.qui.reset()

		if self.inner_editing:
			self.edit_book.close_all_cb()
		else:
			self.edit_book.clear()

		self.res_book.clear()
		self.file_view.clear()
		self.cm_mgr.clear()
		self.code_ctx_view.clear()

	def proj_settings_cb(self):
		proj_args = backend.proj_settings_get()
		proj_args = ProjectUi.show_settings_ui(backend.proj_type(), proj_args)
		if proj_args:
			backend.proj_settings_update(proj_args)

	def editor_tab_changed_cb(self, fname):
		#title = backend.proj_name()
		prj_dir = backend.proj_dir()
		if prj_dir:
			parent = prj_dir
			home_dir = os.path.expanduser('~')
			for i in range(2):
				parent = os.path.dirname(parent)
				if parent == home_dir:
					break
			title = os.path.relpath(prj_dir, parent)
		if not prj_dir:
			title = 'Seascope'
		if fname and fname != '':
			fname = str(fname)
			#if fname.startswith(prj_dir):
				#fname = os.path.relpath(fname, prj_dir)
			title = title + ' - ' + fname
		else:
			fname = 'Seascope'
			title = title + ' - ' + fname
		self.setWindowTitle(title)

	def show_file_line(self, filename, line):
		self.edit_book.show_file_line(filename, line)

	def show_toolbar_cb(self):
		self.is_show_toolbar = self.show_toolbar.isChecked()
		if self.is_show_toolbar:
			self.create_toolbar()
		else:
			self.removeToolBar(self.toolbar)

	def code_ctx_view_act_cb(self):
		self.is_show_code_ctx_view = self.code_ctx_view_act.isChecked()
		if self.is_show_code_ctx_view:
			self.code_ctx_view.show()
		else:
			self.code_ctx_view.hide()

	def connect_signals(self):
		self.edit_book.sig_history_update.connect(self.res_book.history_update)
		self.edit_book.sig_tab_changed.connect(self.editor_tab_changed_cb)
		self.res_book.sig_show_file_line.connect(self.edit_book.show_file_line)
		self.file_view.sig_show_file.connect(self.edit_book.show_file)
		self.edit_book.sig_open_dir_view.connect(self.file_view.open_dir_view)
		self.edit_book.sig_file_closed.connect(self.codemark_del_file_cb)
		self.edit_book.sig_editor_text_selected.connect(self.editor_text_selected)
		self.code_ctx_view.sig_codecontext_showfile.connect(self.code_context_showfile_cb)
		
	def editor_text_selected(self, text):
		if self.is_show_code_ctx_view:
			if not self.code_ctx_view.set_cur_query(text):
				return
			rquery = {}
			rquery['cmd'] = 'DEF'
			rquery['req'] = str(text)
			rquery['opt'] = None
			sig_res = backend.prj.qry.query(rquery)
			sig_res[0].connect(self.code_ctx_view_query_res_cb)

	def code_ctx_view_query_res_cb(self, sym, res):
		self.code_ctx_view.showResult(sym, res)
	
	def code_context_showfile_cb(self, filename, line):
		self.show_file_line(filename, line)

	def setup_widget_tree(self):
		self.hsp = QSplitter();
		self.hsp.addWidget(self.edit_book)
		self.hsp.addWidget(self.file_view)
		self.hsp.setSizes([700, 1])

		self.vsp = QSplitter();
		self.vsp.setOrientation(Qt.Orientation.Vertical)
		self.vsp.addWidget(self.hsp)
		self.vsp.addWidget(self.res_book)
		self.hsp_res = QSplitter();
		self.hsp_res.addWidget(self.res_book)
		self.hsp_res.addWidget(self.code_ctx_view)
		self.vsp.addWidget(self.hsp_res)
		
		self.hsp_res.setSizes([200, 1])
		self.vsp.setSizes([1, 60])
		
		self.setCentralWidget(self.vsp)
		self.setWindowTitle('Seascope')
		self.setGeometry(300, 100, 800, 600)
		#self.showMaximized()

		QApplication.setWindowIcon(QIcon('icons/seascope.svg'))

		# update checked menu item
		self.edit_book.is_show_line = self.eb_is_show_line
		self.edit_book.m_show_line_num.setChecked(self.edit_book.is_show_line)
		if (self.ev_font):
			self.edit_book.ev_font = self.ev_font
			self.code_ctx_view.ev_font = self.ev_font
		self.show_toolbar.setChecked(self.is_show_toolbar)

	def create_widgets(self):
		self.qui_h = QueryUiHelper()
		self.qui = QueryUi(self.qui_h)

		if self.inner_editing:
			self.edit_book = EdViewRW.EditorBookRW()
		else:
			self.edit_book = EdView.EditorBook()

		self.res_book  = ResView.ResultManager()
		self.file_view = FileView.FileTree()
		self.cm_mgr    = CodemarkView.CodemarkManager()
		self.code_ctx_view = CodeContextView.CodeContextViewManager()
		self.code_ctx_view.hide()

		self.sbar = self.statusBar()
		self.create_mbar()
		
		if self.is_show_toolbar:
			self.create_toolbar()

	def setup_widget_hints(self):
		if self.inner_editing:
			EdViewRW.EditorView.ev_popup = self.backend_menu
		else:
			EdView.EditorView.ev_popup = self.backend_menu

		CallView.CallTreeWindow.parent = self
		ClassGraphView.ClassGraphWindow.parent = self
		FileFuncGraphView.FileFuncGraphWindow.parent = self

		self.qui.set_backend_menu(self.backend_menu)
		self.qui_h.edit_book = self.edit_book
		self.qui_h.res_book = self.res_book
		self.qui_h.call_view = CallView
		self.qui_h.class_graph_view = ClassGraphView
		self.qui_h.file_func_graph_view = FileFuncGraphView
		self.qui_h.dbg_view = DebugView
		self.qui_h.file_view = self.file_view

	def setup_style_and_font(self):
		if self.app_style:
			QApplication.setStyle(self.app_style)
		if self.app_font:
			font = QFont()
			font.fromString(self.app_font)
			QApplication.setFont(font)

	def __init__(self, parent=None, app_start_dir=None):
		QMainWindow.__init__(self)
		self.seascope_start_dir = app_start_dir

		self.app_read_config()

		self.create_widgets()
		self.setup_widget_hints()
		self.connect_signals()
		self.setup_widget_tree()
		self.setup_style_and_font()

		prj_path = None
		args = QApplication.arguments()
		if len(args) == 2:
			prj_path = str(args[1])
			if not os.path.isabs(prj_path):
				dname = self.seascope_start_dir
				prj_path = os.path.join(dname, prj_path)
			
		if not prj_path or os.getenv('SEASCOPE_RESTART_HINT'):
			if len(self.recent_projects):
				prj_path = self.recent_projects[0]
		if prj_path:
			self.proj_open(prj_path)
		#else:
			#self.proj_open_cb()
		self.file_restarted_cb()

if __name__ == "__main__":

	# pyqt version 4.5 is required
	pyqt_required_version = 0x40500 
	if not QtCore.PYQT_VERSION >= pyqt_required_version:
		print('Needs pyqt version > 4.5')
		sys.exit(-1)

	# change working dir to the script dir so that we can run this script anywhere else
	app_start_dir = os.getcwd()
	app_dir = os.path.dirname(os.path.realpath(__file__))
	os.chdir(app_dir)
	
	# load plugins
	backend.load_plugins()
	view.load_plugins()

	# start app
	app = QApplication(sys.argv)
	ma = SeascopeApp(app_start_dir=app_start_dir)
	ma.show()
	ret = app.exec()
	sys.exit(ret)



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

__all__ = ["CallView", "ClassGraphView", "FileFuncGraphView", "EdView", "FileView", "ResView", "ProjectUi"]

def load_plugins():
	from . import filecontext
	filecontext.load_plugins()